#ifndef _GSH_OPENGL_H_
#define _GSH_OPENGL_H_

#include "GSHandler.h"
#include "win32/Window.h"
#include <gl/gl.h>

#define PREF_CGSH_OPENGL_LINEASQUADS "renderer.opengl.linesasquads"

class CGSH_OpenGL : public CGSHandler, public CSettingsDialogProvider
{
public:
									CGSH_OpenGL(Framework::CWindow*);
									~CGSH_OpenGL();

	static void						CreateGSHandler(Framework::CWindow*);

	void							WriteRegister(uint8, uint64);
	void							ProcessImageTransfer(uint32, uint32);

	void							Flip();
	CSettingsDialogProvider*		GetSettingsDialogProvider();

	CModalWindow*					CreateSettingsDialog(HWND);
	void							OnSettingsDialogDestroyed();

	bool							IsColorTableExtSupported();
	bool							IsBlendColorExtSupported();
	bool							IsBlendEquationExtSupported();
	bool							IsRGBA5551ExtSupported();

private:
	enum MAXCACHE
	{
		MAXCACHE = 5,
	};

	enum CVTBUFFERSIZE
	{
		CVTBUFFERSIZE = 0x200000,
	};

	typedef void (CGSH_OpenGL::*TEXTUREUPLOADER)(GSTEX0*, GSTEXA*);

	struct VERTEX
	{
		uint64						nPosition;
		uint64						nRGBAQ;
		uint64						nUV;
		uint64						nST;
	};

	class CTexture
	{
	public:
									CTexture();
									~CTexture();
		void						Invalidate();
		void						InvalidateFromMemorySpace(uint32, uint32);
		bool						IsValid();

		uint32						m_nStart;
		uint32						m_nSize;
		unsigned int				m_nTexture;
	};

	static CGSHandler*				GSHandlerFactory(void*);

	void							InitializeRC();
	void							SetViewport(int, int);
	void							SetReadCircuitMatrix(int, int);
	void							LinearZOrtho(double, double, double, double);
	void							UpdateViewport();
	unsigned int					LoadTexture(GSTEX0*, GSTEX1*);

	double							GetZ(double);
	unsigned int					GetNextPowerOf2(unsigned int);

	void							Prim_Point();
	void							Prim_Line();
	void							Prim_Triangle();
	void							Prim_Sprite();

	void							SetRenderingContext(unsigned int);
	void							SetupTestFunctions(uint64);
	void							SetupDepthBuffer(uint64);
	void							SetupBlendingFunction(uint64);
	void							SetupTexture(uint64, uint64);

	void							DisplayTransferedImage(uint32);

	void							TexUploader_Psm8_Cvt(GSTEX0*, GSTEXA*);
	void							TexUploader_Psm8_Hw(GSTEX0*, GSTEXA*);

	void							TexUploader_Psm16_Cvt(GSTEX0*, GSTEXA*);
	void							TexUploader_Psm16_Hw(GSTEX0*, GSTEXA*);

	//Context variables (put this in a struct or something?)
	double							m_nPrimOfsX;
	double							m_nPrimOfsY;
	uint32							m_nTexWidth;
	uint32							m_nTexHeight;
	unsigned int					m_nTexHandle;
	double							m_nMaxZ;

	int								m_nWidth;
	int								m_nHeight;

	bool							m_nLinesAsQuads;

	uint8*							m_pCvtBuffer;

	void							VerifyRGBA5551Support();
	bool							m_nIsRGBA5551Supported;

	unsigned int					TexCache_Search(uint32);
	void							TexCache_Insert(uint32, uint32, unsigned int);
	void							TexCache_InvalidateTextures(uint32, uint32);

	CTexture						m_TexCache[MAXCACHE];
	unsigned int					m_nTexCacheIndex;

	VERTEX							m_VtxBuffer[3];
	int								m_nVtxCount;

	GSPRMODE						m_PrimitiveMode;
	unsigned int					m_nPrimitiveType;

	TEXTUREUPLOADER					m_pTexUploader_Psm8;
	TEXTUREUPLOADER					m_pTexUploader_Psm16;

	Framework::CWindow*				m_pOutputWnd;

	HGLRC							m_hRC;
	HDC								m_hDC;
	static PIXELFORMATDESCRIPTOR	m_PFD;

};

#endif