#include <assert.h>
#include <stdio.h>
#include "IOP_Cdvdfsv.h"
#include "PS2VM.h"

using namespace IOP;
using namespace Framework;

uint32	CCdvdfsv::m_nStreamPos = 0;

//593 - Method 0x22 CdMmode

CCdvdfsv::CCdvdfsv(uint32 nID)
{
	m_nID = nID;
}

CCdvdfsv::~CCdvdfsv()
{

}

void CCdvdfsv::Invoke(uint32 nMethod, void* pArgs, uint32 nArgsSize, void* pRet, uint32 nRetSize)
{
	switch(m_nID)
	{
	case MODULE_ID_1:
		Invoke592(nMethod, pArgs, nArgsSize, pRet, nRetSize);
		break;
	case MODULE_ID_4:
		Invoke595(nMethod, pArgs, nArgsSize, pRet, nRetSize);
		break;
	case MODULE_ID_6:
		Invoke597(nMethod, pArgs, nArgsSize, pRet, nRetSize);
		break;
	}
}

void CCdvdfsv::SaveState(CStream* pStream)
{
	if(m_nID != MODULE_ID_1) return;

	pStream->Write(&m_nStreamPos, 4);
}

void CCdvdfsv::LoadState(CStream* pStream)
{
	if(m_nID != MODULE_ID_1) return;

	pStream->Read(&m_nStreamPos, 4);
}

void CCdvdfsv::Invoke592(uint32 nMethod, void* pArgs, uint32 nArgsSize, void* pRet, uint32 nRetSize)
{
	switch(nMethod)
	{
	case 0:
		//Init
		break;
	}
}

void CCdvdfsv::Invoke595(uint32 nMethod, void* pArgs, uint32 nArgsSize, void* pRet, uint32 nRetSize)
{
	switch(nMethod)
	{
	case 9:
		StreamCmd(pArgs, nArgsSize, pRet, nRetSize);
		break;
	}
}

void CCdvdfsv::Invoke597(uint32 nMethod, void* pArgs, uint32 nArgsSize, void* pRet, uint32 nRetSize)
{
	switch(nMethod)
	{
	case 0:
		SearchFile(pArgs, nArgsSize, pRet, nRetSize);
		break;
	default:
		printf("IOP_Cdvdfsv: Unknown method called (0x%0.8X, 0x%0.8X).\r\n", m_nID, nMethod);
		break;
	}
}

void CCdvdfsv::StreamCmd(void* pArgs, uint32 nArgsSize, void* pRet, uint32 nRetSize)
{
	uint32 nSector, nCount, nDstAddr, nCmd, nMode, i;

	nSector		= ((uint32*)pArgs)[0x00];
	nCount		= ((uint32*)pArgs)[0x01];
	nDstAddr	= ((uint32*)pArgs)[0x02];
	nCmd		= ((uint32*)pArgs)[0x03];
	nMode		= ((uint32*)pArgs)[0x04];

	printf("IOP_Cdvdfsv: StreamCmd(sector = 0x%0.8X, count = 0x%0.8X, addr = 0x%0.8X, cmd = 0x%0.8X);\r\n", \
		nSector,
		nCount,
		nDstAddr,
		nMode);

	switch(nCmd)
	{
	case 1:
		//Start
		m_nStreamPos = nSector;
		((uint32*)pRet)[0] = 1;
		
		printf("IOP_Cdvdfsv: StreamStart(pos = 0x%0.8X);\r\n", \
			nSector);
		break;
	case 2:
		//Read
		nDstAddr &= (CPS2VM::RAMSIZE - 1);

		for(i = 0; i < nCount; i++)
		{
			CPS2VM::m_pCDROM0->ReadBlock(m_nStreamPos, CPS2VM::m_pRAM + (nDstAddr + (i * 0x800)));
			m_nStreamPos++;
		}

		((uint32*)pRet)[0] = nCount;

		printf("IOP_Cdvdfsv: StreamRead(count = 0x%0.8X, dest = 0x%0.8X);\r\n", \
			nCount, \
			nDstAddr);
		break;
	case 5:
		//Init
		((uint32*)pRet)[0] = 1;

		printf("IOP_Cdvdfsv: StreamInit(bufsize = 0x%0.8X, numbuf = 0x%0.8X, buf = 0x%0.8X);\r\n", \
			nSector, \
			nCount, \
			nDstAddr);
		break;
	case 4:
	case 9:
		//Seek
		m_nStreamPos = nSector;
		((uint32*)pRet)[0] = 1;

		printf("IOP_Cdvdfsv: StreamSeek(pos = 0x%0.8X);\r\n", \
			nSector);
		break;
	}
}

void CCdvdfsv::SearchFile(void* pArgs, uint32 nArgsSize, void* pRet, uint32 nRetSize)
{
	char* sPath;
	CISO9660* pISO;
	ISO9660::CDirectoryRecord Record;

	assert(nArgsSize == 0x12C);
	assert(nRetSize == 4);

	pISO = CPS2VM::m_pCDROM0;
	if(pISO == NULL)
	{
		*(uint32*)pRet = 0;
		return;
	}

	//0x12C structure
	//00 - Block Num
	//04 - Size
	//08
	//0C
	//10
	//14
	//18
	//1C
	//20 - Unknown
	//24 - Path

	sPath = (char*)pArgs + 0x24;

	printf("IOP_Cdvdfsv: SearchFile(path = %s);\r\n", sPath);

	if(!pISO->GetFileRecord(&Record, sPath))
	{
		*(uint32*)pRet = 0;
		return;
	}

	((uint32*)pArgs)[0x00] = Record.GetPosition();
	((uint32*)pArgs)[0x01] = Record.GetDataLength();

	*(uint32*)pRet = 1;
}
