#ifndef _MA_VU_H_
#define _MA_VU_H_

#include "MIPSArchitecture.h"
#include "MIPSReflection.h"

class CMA_VU : public CMIPSArchitecture
{
public:
											CMA_VU();
	virtual void							CompileInstruction(uint32, CCacheBlock*, CMIPS*, bool);
	virtual void							GetInstructionMnemonic(CMIPS*, uint32, uint32, char*, unsigned int);
	virtual void							GetInstructionOperands(CMIPS*, uint32, uint32, char*, unsigned int);
	virtual bool							IsInstructionBranch(CMIPS*, uint32, uint32);
	virtual uint32							GetInstructionEffectiveAddress(CMIPS*, uint32, uint32);

private:
	void									SetupReflectionTables();

	class CUpper
	{
	public:
		void								SetupReflectionTables();
		void								CompileInstruction(uint32, CCacheBlock*, CMIPS*, bool);
		void								GetInstructionMnemonic(CMIPS*, uint32, uint32, char*, unsigned int);
		void								GetInstructionOperands(CMIPS*, uint32, uint32, char*, unsigned int);
		bool								IsInstructionBranch(CMIPS*, uint32, uint32);
		uint32								GetInstructionEffectiveAddress(CMIPS*, uint32, uint32);
	
	private:
		static void							(*m_pOpVector[0x40])();
		static void							(*m_pOpVector0[0x20])();
		static void							(*m_pOpVector1[0x20])();
		static void							(*m_pOpVector2[0x20])();
		static void							(*m_pOpVector3[0x20])();

		static uint32						m_nOpcode;
		static uint8						m_nFT;
		static uint8						m_nFS;
		static uint8						m_nFD;
		static uint8						m_nBc;
		static uint8						m_nDest;

		static void							ReflOpFtFs(MIPSReflection::INSTRUCTION*, CMIPS*, uint32, uint32, char*, unsigned int);

		static void							LOI(uint32);

		//Vector
		static void							ADDbc();
		static void							SUBbc();
		static void							MADDbc();
		static void							MAXbc();
		static void							MULbc();
		static void							MULq();
		static void							MULi();
		static void							SUBi();
		static void							MSUBi();
		static void							ADD();
		static void							MADD();
		static void							MUL();
		static void							SUB();
		static void							VECTOR0();
		static void							VECTOR1();
		static void							VECTOR2();
		static void							VECTOR3();

		//Vector X Common
		static void							MADDAbc();
		static void							MULAbc();

		//Vector 0
		static void							ITOF0();
		static void							FTOI0();

		//Vector 1
		static void							FTOI4();
		static void							ABS();
		static void							MADDA();

		//Vector 2
		static void							MULAi();
		static void							MULA();

		//Vector 3
		static void							MADDAi();
		static void							MSUBAi();
		static void							NOP();

		MIPSReflection::INSTRUCTION			m_ReflV[64];
		MIPSReflection::INSTRUCTION			m_ReflVX0[32];
		MIPSReflection::INSTRUCTION			m_ReflVX1[32];
		MIPSReflection::INSTRUCTION			m_ReflVX2[32];
		MIPSReflection::INSTRUCTION			m_ReflVX3[32];

		MIPSReflection::SUBTABLE			m_ReflVTable;
		MIPSReflection::SUBTABLE			m_ReflVX0Table;
		MIPSReflection::SUBTABLE			m_ReflVX1Table;
		MIPSReflection::SUBTABLE			m_ReflVX2Table;
		MIPSReflection::SUBTABLE			m_ReflVX3Table;

		static MIPSReflection::INSTRUCTION	m_cReflV[64];
		static MIPSReflection::INSTRUCTION	m_cReflVX0[32];
		static MIPSReflection::INSTRUCTION	m_cReflVX1[32];
		static MIPSReflection::INSTRUCTION	m_cReflVX2[32];
		static MIPSReflection::INSTRUCTION	m_cReflVX3[32];
	};

	class CLower
	{
	public:
		void								SetupReflectionTables();
		void								CompileInstruction(uint32, CCacheBlock*, CMIPS*, bool);
		void								GetInstructionMnemonic(CMIPS*, uint32, uint32, char*, unsigned int);
		void								GetInstructionOperands(CMIPS*, uint32, uint32, char*, unsigned int);
		bool								IsInstructionBranch(CMIPS*, uint32, uint32);
		uint32								GetInstructionEffectiveAddress(CMIPS*, uint32, uint32);

	private:
		static void							(*m_pOpGeneral[0x80])();
		static void							(*m_pOpLower[0x40])();
		static void							(*m_pOpVector0[0x20])();
		static void							(*m_pOpVector1[0x20])();
		static void							(*m_pOpVector3[0x20])();

		static uint32						m_nOpcode;
		static uint32						m_nAddress;
		static uint8						m_nIT;
		static uint8						m_nIS;
		static uint8						m_nID;
		static uint8						m_nFSF;
		static uint8						m_nFTF;
		static uint8						m_nDest;
		static uint16						m_nImm11;
		static uint16						m_nImm15;
		static uint16						m_nImm15S;

		static uint32						GetDestOffset(uint8);
		static void							SetBranchAddress(bool, int32);

		static void							ReflOpIs(MIPSReflection::INSTRUCTION*, CMIPS*, uint32, uint32, char*, unsigned int);
		static void							ReflOpIsOfs(MIPSReflection::INSTRUCTION*, CMIPS*, uint32, uint32, char*, unsigned int);
		static void							ReflOpIt(MIPSReflection::INSTRUCTION*, CMIPS*, uint32, uint32, char*, unsigned int);
		static void							ReflOpItIs(MIPSReflection::INSTRUCTION*, CMIPS*, uint32, uint32, char*, unsigned int);
		static void							ReflOpItFsf(MIPSReflection::INSTRUCTION*, CMIPS*, uint32, uint32, char*, unsigned int);
		static void							ReflOpItIsOfs(MIPSReflection::INSTRUCTION*, CMIPS*, uint32, uint32, char*, unsigned int);
		static void							ReflOpItIsImm15(MIPSReflection::INSTRUCTION*, CMIPS*, uint32, uint32, char*, unsigned int);
		static void							ReflOpIdIsIt(MIPSReflection::INSTRUCTION*, CMIPS*, uint32, uint32, char*, unsigned int);
		static void							ReflOpItOfsIsDst(MIPSReflection::INSTRUCTION*, CMIPS*, uint32, uint32, char*, unsigned int);
		static void							ReflOpFtDstIsInc(MIPSReflection::INSTRUCTION*, CMIPS*, uint32, uint32, char*, unsigned int);
		static void							ReflOpFtDstIsDec(MIPSReflection::INSTRUCTION*, CMIPS*, uint32, uint32, char*, unsigned int);
		static void							ReflOpFsDstItInc(MIPSReflection::INSTRUCTION*, CMIPS*, uint32, uint32, char*, unsigned int);
		static void							ReflOpFtDstOfsIs(MIPSReflection::INSTRUCTION*, CMIPS*, uint32, uint32, char*, unsigned int);
		static void							ReflOpFtDstFsDst(MIPSReflection::INSTRUCTION*, CMIPS*, uint32, uint32, char*, unsigned int);

		static uint32						ReflEaOffset(MIPSReflection::INSTRUCTION*, CMIPS*, uint32, uint32);

		//General
		static void							LQ();
		static void							ILW();
		static void							ISW();
		static void							IADDIU();
		static void							ISUBIU();
		static void							FMAND();
		static void							IBEQ();
		static void							IBNE();
		static void							IBGEZ();
		static void							LOWEROP();

		//LowerOp
		static void							IADD();
		static void							IOR();
		static void							VECTOR0();
		static void							VECTOR1();
		static void							VECTOR3();

		//Vector0
		static void							LQI();
		static void							DIV();
		static void							MTIR();
		static void							XTOP();
		static void							XGKICK();

		//Vector1
		static void							SQI();

		//Vector3
		static void							WAITQ();

		MIPSReflection::INSTRUCTION			m_ReflGeneral[128];
		MIPSReflection::INSTRUCTION			m_ReflV[64];
		MIPSReflection::INSTRUCTION			m_ReflVX0[32];
		MIPSReflection::INSTRUCTION			m_ReflVX1[32];
		MIPSReflection::INSTRUCTION			m_ReflVX3[32];

		MIPSReflection::SUBTABLE			m_ReflGeneralTable;
		MIPSReflection::SUBTABLE			m_ReflVTable;
		MIPSReflection::SUBTABLE			m_ReflVX0Table;
		MIPSReflection::SUBTABLE			m_ReflVX1Table;
		MIPSReflection::SUBTABLE			m_ReflVX3Table;

		static MIPSReflection::INSTRUCTION	m_cReflGeneral[128];
		static MIPSReflection::INSTRUCTION	m_cReflV[64];
		static MIPSReflection::INSTRUCTION	m_cReflVX0[32];
		static MIPSReflection::INSTRUCTION	m_cReflVX1[32];
		static MIPSReflection::INSTRUCTION	m_cReflVX3[32];
	};

	CUpper									m_Upper;
	CLower									m_Lower;
};

extern CMA_VU g_MAVU;

#endif
