#include "MA_VU.h"
#include "MIPS.h"
#include "VIF.h"
#include "VUShared.h"

uint32			CMA_VU::CLower::m_nOpcode;
uint16			CMA_VU::CLower::m_nImm11;
uint16			CMA_VU::CLower::m_nImm15;
uint16			CMA_VU::CLower::m_nImm15S;
uint8			CMA_VU::CLower::m_nIT;
uint8			CMA_VU::CLower::m_nIS;
uint8			CMA_VU::CLower::m_nID;
uint8			CMA_VU::CLower::m_nFSF;
uint8			CMA_VU::CLower::m_nFTF;
uint8			CMA_VU::CLower::m_nDest;

void CMA_VU::CLower::CompileInstruction(uint32 nAddress, CCacheBlock* pBlock, CMIPS* pCtx, bool nParent)
{
	uint32 nPrevOpcode, nNextOpcode;

	m_nOpcode	= pCtx->m_pMemoryMap->GetWord(nAddress);
	nPrevOpcode	= pCtx->m_pMemoryMap->GetWord(nAddress - 8);
	nNextOpcode = pCtx->m_pMemoryMap->GetWord(nAddress + 4);

	if(IsInstructionBranch(pCtx, nAddress - 8, nPrevOpcode))
	{
		m_pB->SetDelayedJumpCheck();
	}

	if(nNextOpcode & 0x80000000)
	{
//		LOI();
		return;
	}

	m_pB		= pBlock;
	m_pCtx		= pCtx;

	m_nDest		= (uint8 )((m_nOpcode >> 21) & 0x000F);

	m_nFSF		= ((m_nDest >> 0) & 0x03);
	m_nFTF		= ((m_nDest >> 2) & 0x03);

	m_nIT		= (uint8 )((m_nOpcode >> 16) & 0x001F);
	m_nIS		= (uint8 )((m_nOpcode >> 11) & 0x001F);
	m_nID		= (uint8 )((m_nOpcode >>  6) & 0x001F);
	m_nImm11	= (uint16)((m_nOpcode >>  0) & 0x07FF);
	m_nImm15	= (uint16)((m_nOpcode & 0x7FF) | (m_nOpcode & 0x01E00000) >> 10);
	m_nImm15S	= m_nImm15 | (m_nImm15 & 0x4000 ? 0x8000 : 0x0000);

	if(m_nOpcode != 0x8000033C)
	{
		m_pOpGeneral[m_nOpcode >> 25]();
	}
}

void CMA_VU::CLower::SetBranchAddress(bool nCondition, int32 nOffset)
{
	m_pB->PushImm(MIPS_INVALID_PC);
	m_pB->PullAddr(&m_pCtx->m_State.nDelayedJumpAddr);

	m_pB->BeginJcc(nCondition);
	{
		m_pB->PushAddr(&m_pCtx->m_State.nPC);
		m_pB->AddImm(nOffset);
		m_pB->PullAddr(&m_pCtx->m_State.nDelayedJumpAddr);
	}
	m_pB->EndJcc();
}

uint32 CMA_VU::CLower::GetDestOffset(uint8 nDest)
{
	if(nDest & 0x0001) return 0xC;
	if(nDest & 0x0002) return 0x8;
	if(nDest & 0x0004) return 0x4;
	if(nDest & 0x0008) return 0x0;

	return 0;
}

//////////////////////////////////////////////////
//General Instructions
//////////////////////////////////////////////////

//00
void CMA_VU::CLower::LQ()
{
	unsigned int i;

	for(i = 0; i < 4; i++)
	{
		if(!VUShared::DestinationHasElement(m_nDest, i)) continue;

		m_pB->PushAddr(&m_pCtx->m_State.nCOP2VI[m_nIS]);
		m_pB->AddImm((uint32)VUShared::GetImm11Offset(m_nImm11));
		m_pB->SllImm(4);
		m_pB->AddImm(i * 4);

		m_pB->PushRef(m_pCtx);
		m_pB->Call(&CCacheBlock::GetWordProxy, 2, true);

		m_pB->PullAddr(VUShared::GetVectorElement(m_pCtx, m_nIT, i));
	}
}

//04
void CMA_VU::CLower::ILW()
{
	m_pB->PushAddr(&m_pCtx->m_State.nCOP2VI[m_nIS]);
	m_pB->AddImm((uint32)VUShared::GetImm11Offset(m_nImm11));

	m_pB->SllImm(4);
	m_pB->AddImm(GetDestOffset(m_nDest));

	m_pB->PushRef(m_pCtx);
	m_pB->Call(&CCacheBlock::GetWordProxy, 2, true);
	m_pB->PullAddr(&m_pCtx->m_State.nCOP2VI[m_nIT]);
}

//05
void CMA_VU::CLower::ISW()
{
	m_pB->PushAddr(&m_pCtx->m_State.nCOP2VI[m_nIS]);
	m_pB->AddImm((uint32)VUShared::GetImm11Offset(m_nImm11));

	m_pB->SllImm(4);
	m_pB->AddImm(GetDestOffset(m_nDest));

	m_pB->PushAddr(&m_pCtx->m_State.nCOP2VI[m_nIT]);
	m_pB->AndImm(0xFFFF);

	m_pB->PushRef(m_pCtx);
	m_pB->Call(&CCacheBlock::SetWordProxy, 3, false);
}

//08
void CMA_VU::CLower::IADDIU()
{
	m_pB->PushAddr(&m_pCtx->m_State.nCOP2VI[m_nIS]);
	m_pB->AddImm((uint16)m_nImm15);
	m_pB->PullAddr(&m_pCtx->m_State.nCOP2VI[m_nIT]);
}

//09
void CMA_VU::CLower::ISUBIU()
{
	m_pB->PushAddr(&m_pCtx->m_State.nCOP2VI[m_nIS]);
	m_pB->SubImm((uint16)m_nImm15);
	m_pB->PullAddr(&m_pCtx->m_State.nCOP2VI[m_nIT]);
}

//1A
void CMA_VU::CLower::FMAND()
{
	unsigned int i;

	//MAC flag temp
	m_pB->PushImm(0);

	//Check sign flags
	for(i = 0; i < 4; i++)
	{
		m_pB->PushAddr(&m_pCtx->m_State.nCOP2SF.nV[3 - i]);
		m_pB->PushImm(0);
		m_pB->Cmp(JCC_CONDITION_NE);
		m_pB->SllImm(4 + i);
		m_pB->Or();
	}

	m_pB->PushAddr(&m_pCtx->m_State.nCOP2VI[m_nIS]);
	m_pB->And();
	m_pB->PullAddr(&m_pCtx->m_State.nCOP2VI[m_nIT]);
}

//28
void CMA_VU::CLower::IBEQ()
{
	m_pB->PushAddr(&m_pCtx->m_State.nCOP2VI[m_nIS]);
	m_pB->AndImm(0xFFFF);
	m_pB->PushAddr(&m_pCtx->m_State.nCOP2VI[m_nIT]);
	m_pB->AndImm(0xFFFF);

	m_pB->Cmp(JCC_CONDITION_EQ);

	SetBranchAddress(true, VUShared::GetBranch(m_nImm11) + 4);
}

//29
void CMA_VU::CLower::IBNE()
{
	m_pB->PushAddr(&m_pCtx->m_State.nCOP2VI[m_nIS]);
	m_pB->AndImm(0xFFFF);
	m_pB->PushAddr(&m_pCtx->m_State.nCOP2VI[m_nIT]);
	m_pB->AndImm(0xFFFF);

	m_pB->Cmp(JCC_CONDITION_EQ);

	SetBranchAddress(false, VUShared::GetBranch(m_nImm11) + 4);
}

//2F
void CMA_VU::CLower::IBGEZ()
{
	m_pB->PushImm(0);
	m_pB->PushAddr(&m_pCtx->m_State.nCOP2VI[m_nIS]);
	m_pB->AndImm(0x8000);

	m_pB->Cmp(JCC_CONDITION_EQ);

	SetBranchAddress(true, VUShared::GetBranch(m_nImm11) + 4);
}

//40
void CMA_VU::CLower::LOWEROP()
{
	m_pOpLower[m_nOpcode & 0x3F]();
}

//////////////////////////////////////////////////
//LowerOp Instructions
//////////////////////////////////////////////////

//30
void CMA_VU::CLower::IADD()
{
	m_pB->PushAddr(&m_pCtx->m_State.nCOP2VI[m_nIS]);
	m_pB->PushAddr(&m_pCtx->m_State.nCOP2VI[m_nIT]);
	m_pB->Add();
	m_pB->PullAddr(&m_pCtx->m_State.nCOP2VI[m_nID]);
}

//35
void CMA_VU::CLower::IOR()
{
	m_pB->PushAddr(&m_pCtx->m_State.nCOP2VI[m_nIS]);
	m_pB->PushAddr(&m_pCtx->m_State.nCOP2VI[m_nIT]);
	m_pB->Or();
	m_pB->PullAddr(&m_pCtx->m_State.nCOP2VI[m_nID]);
}

//3C
void CMA_VU::CLower::VECTOR0()
{
	m_pOpVector0[(m_nOpcode >> 6) & 0x1F]();
}

//3D
void CMA_VU::CLower::VECTOR1()
{
	m_pOpVector1[(m_nOpcode >> 6) & 0x1F]();
}

//3F
void CMA_VU::CLower::VECTOR3()
{
	m_pOpVector3[(m_nOpcode >> 6) & 0x1F]();
}

//////////////////////////////////////////////////
//Vector0 Instructions
//////////////////////////////////////////////////

//0D
void CMA_VU::CLower::LQI()
{
	unsigned int i;

	for(i = 0; i < 4; i++)
	{
		if(!VUShared::DestinationHasElement(m_nDest, i)) continue;

		m_pB->PushAddr(&m_pCtx->m_State.nCOP2VI[m_nIS]);
		m_pB->SllImm(4);
		m_pB->AddImm(i * 4);

		m_pB->PushRef(m_pCtx);
		m_pB->Call(&CCacheBlock::GetWordProxy, 2, true);

		m_pB->PullAddr(VUShared::GetVectorElement(m_pCtx, m_nIT, i));
	}

	m_pB->PushAddr(&m_pCtx->m_State.nCOP2VI[m_nIS]);
	m_pB->AddImm(1);
	m_pB->PullAddr(&m_pCtx->m_State.nCOP2VI[m_nIS]);
}

//0E
void CMA_VU::CLower::DIV()
{
	VUShared::DIV(m_pB, m_pCtx, m_nIS, m_nFSF, m_nIT, m_nFTF);
}

//0F
void CMA_VU::CLower::MTIR()
{
	m_pB->PushAddr(&m_pCtx->m_State.nCOP2[m_nIS].nV[m_nFSF]);
	m_pB->PullAddr(&m_pCtx->m_State.nCOP2VI[m_nIT]);
}

//1A
void CMA_VU::CLower::XTOP()
{
	m_pB->PushAddr(CVIF::GetTop1Address());
	m_pB->PullAddr(&m_pCtx->m_State.nCOP2VI[m_nIT]);
}

//1B
void CMA_VU::CLower::XGKICK()
{
	m_pB->PushAddr(&m_pCtx->m_State.nCOP2VI[m_nIS]);
	m_pB->Call(CVIF::ProcessXGKICK, 1, false);
}

//////////////////////////////////////////////////
//Vector1 Instructions
//////////////////////////////////////////////////

//0D
void CMA_VU::CLower::SQI()
{
	unsigned int i;

	for(i = 0; i < 4; i++)
	{
		if(!VUShared::DestinationHasElement(m_nDest, i)) continue;

		m_pB->PushAddr(&m_pCtx->m_State.nCOP2VI[m_nIT]);
		m_pB->SllImm(4);
		m_pB->AddImm(i * 4);

		m_pB->PushAddr(VUShared::GetVectorElement(m_pCtx, m_nIS, i));

		m_pB->PushRef(m_pCtx);
		m_pB->Call(&CCacheBlock::SetWordProxy, 3, false);
	}

	m_pB->PushAddr(&m_pCtx->m_State.nCOP2VI[m_nIT]);
	m_pB->AddImm(1);
	m_pB->PullAddr(&m_pCtx->m_State.nCOP2VI[m_nIT]);
}

//////////////////////////////////////////////////
//Vector3 Instructions
//////////////////////////////////////////////////

//0E
void CMA_VU::CLower::WAITQ()
{

}

//////////////////////////////////////////////////
//Opcode Tables
//////////////////////////////////////////////////

void (*CMA_VU::CLower::m_pOpGeneral[0x80])() =
{
	//0x00
	LQ,				Illegal,		Illegal,		Illegal,		ILW,			ISW,			Illegal,		Illegal,
	//0x08
	IADDIU,			ISUBIU,			Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x10
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x18
	Illegal,		Illegal,		FMAND,			Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x20
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x28
	IBEQ,			IBNE,			Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		IBGEZ,
	//0x30
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x38
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x40
	LOWEROP,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x48
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x50
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x58
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x60
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x68
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x70
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x78
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
};

void (*CMA_VU::CLower::m_pOpLower[0x40])() =
{
	//0x00
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x08
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x10
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x18
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x20
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x28
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x30
	IADD,			Illegal,		Illegal,		Illegal,		Illegal,		IOR,			Illegal,		Illegal,
	//0x38
	Illegal,		Illegal,		Illegal,		Illegal,		VECTOR0,		VECTOR1,		Illegal,		VECTOR3,
};

void (*CMA_VU::CLower::m_pOpVector0[0x20])() =
{
	//0x00
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x08
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		LQI,			DIV,			MTIR,
	//0x10
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x18
	Illegal,		Illegal,		XTOP,			XGKICK,			Illegal,		Illegal,		Illegal,		Illegal,
};

void (*CMA_VU::CLower::m_pOpVector1[0x20])() =
{
	//0x00
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x08
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		SQI,			Illegal,		Illegal,
	//0x10
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x18
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
};

void (*CMA_VU::CLower::m_pOpVector3[0x20])() =
{
	//0x00
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x08
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		WAITQ,			Illegal,
	//0x10
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
	//0x18
	Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,		Illegal,
};
