#include <assert.h>
#include "VIF.h"
#include "GIF.h"
#include "PS2VM.h"

using namespace Framework;

uint32		CVIF::m_VPU_STAT;

uint32		CVIF::m_VIF0_CYCLE;

uint32		CVIF::m_VIF1_STAT;
uint32		CVIF::m_VIF1_CYCLE;
uint32		CVIF::m_VIF1_BASE;
uint32		CVIF::m_VIF1_TOP;
uint32		CVIF::m_VIF1_TOPS;
uint32		CVIF::m_VIF1_OFST;

void CVIF::Reset()
{
	m_VPU_STAT = 0;
	m_VIF1_STAT = 0;
}

void CVIF::SaveState(CStream* pStream)
{
	pStream->Write(&m_VPU_STAT,		4);
	pStream->Write(&m_VIF1_STAT,	4);
	pStream->Write(&m_VIF1_CYCLE,	4);
	pStream->Write(&m_VIF1_BASE,	4);
	pStream->Write(&m_VIF1_TOP,		4);
	pStream->Write(&m_VIF1_TOPS,	4);
	pStream->Write(&m_VIF1_OFST,	4);
}

void CVIF::LoadState(CStream* pStream)
{
	pStream->Read(&m_VPU_STAT,		4);
	pStream->Read(&m_VIF1_STAT,		4);
	pStream->Read(&m_VIF1_CYCLE,	4);
	pStream->Read(&m_VIF1_BASE,		4);
	pStream->Read(&m_VIF1_TOP,		4);
	pStream->Read(&m_VIF1_TOPS,		4);
	pStream->Read(&m_VIF1_OFST,		4);
}

uint32 CVIF::ReceiveDMA1(uint32 nAddress, uint32 nQWC)
{
	ProcessPacket(nAddress, nQWC * 0x10, 1);
	return nQWC;
}

uint32* CVIF::GetTop1Address()
{
	return &m_VIF1_TOP;
}

void CVIF::StopVU(CMIPS* pCtx)
{
	if(pCtx == &CPS2VM::m_VU1)
	{
		m_VPU_STAT &= ~STAT_VBS1;
	}
}

void CVIF::ProcessXGKICK(uint32 nAddress)
{
	nAddress &= 0xFFFF;
	nAddress *= 0x10;

	CGIF::ProcessPacket(CPS2VM::m_pVUMem1, nAddress, CPS2VM::VUMEM1SIZE);
}

bool CVIF::IsVU1Running()
{
	return (m_VPU_STAT & STAT_VBS1) != 0;
}

void CVIF::ProcessPacket(uint32 nAddress, uint32 nSize, unsigned int nUnit)
{
	uint32 nCode, nEnd;
	uint32 nImm, nNum;
	uint8 nCmd, nWL, nCL;
	bool nUsn, nFlg;
	uint8* pVUMemory;
	uint8* pMicroMemory;
	uint32* pCycle;

	nEnd = nAddress + nSize;

	pVUMemory		= (nUnit == 0) ? CPS2VM::m_pVUMem0 : CPS2VM::m_pVUMem1;
	pMicroMemory	= (nUnit == 0) ? CPS2VM::m_pMicroMem0 : CPS2VM::m_pMicroMem1;
	pCycle			= (nUnit == 0) ? &m_VIF0_CYCLE : &m_VIF1_CYCLE;

	while(nAddress < nEnd)
	{
		nCode = *(uint32*)&CPS2VM::m_pRAM[nAddress];

		nAddress	+= 4;

		nCmd	= (uint8 )(nCode >> 24) & 0x00FF;
		nNum	= (uint32)(nCode >> 16) & 0x00FF;
		nImm	= (uint32)(nCode >>  0) & 0xFFFF;

		if(nCmd & 0x80)
		{
			assert(0);
		}
		nCmd &= 0x7F;

		if(nCmd >= 0x60)
		{
			//UNPACK
			nUsn = (nImm & 0x4000) != 0;
			nFlg = (nImm & 0x8000) != 0;
			nImm = (nImm & 0x03FF);

			if(nFlg == 1 && nUnit == 1)
			{
				nImm += (uint16)m_VIF1_TOPS;
			}

			nImm *= 0x10;
			nNum *= 0x10;

			nWL = (uint8)(((*pCycle) & 0xFF00) >> 8);
			nCL = (uint8)(((*pCycle) & 0x00FF) >> 0);

			//Check format
			switch(nCmd & 0x0F)
			{
			case 0x0C:
				//V4-32
				nAddress += UnpackV432(nAddress, nImm, nNum, nUsn, nWL, nCL, pVUMemory);
				break;
			case 0x0F:
				//V4-5
				nAddress += UnpackV45(nAddress, nImm, nNum, nUsn, nWL, nCL, pVUMemory);
				break;
			default:
				assert(0);
				break;
			}
		}
		else
		{
			switch(nCmd)
			{
			case 0x00:
				//NOP
				break;
			case 0x01:
				//STCYCL
				if(nUnit == 0) m_VIF0_CYCLE = nImm;
				if(nUnit == 1) m_VIF1_CYCLE = nImm;
				break;
			case 0x02:
				//OFFSET
				if(nUnit == 0)
				{
					assert(0);
					break;
				}

				m_VIF1_OFST = nImm;
				
				m_VIF1_STAT &= ~STAT1_DBF;
				m_VIF1_TOPS = m_VIF1_BASE;
				break;
			case 0x03:
				//BASE
				if(nUnit == 0) assert(0);
				if(nUnit == 1) m_VIF1_BASE = nImm;
				break;
			case 0x11:
				//FLUSH
				break;
			case 0x14:
				//MSCAL
				if(nUnit == 0) assert(0);

				//Setup the TOP address
				m_VIF1_TOP = m_VIF1_TOPS;

				if(m_VIF1_STAT & STAT1_DBF)
				{
					m_VIF1_TOPS = m_VIF1_BASE;
				}
				else
				{
					m_VIF1_TOPS = m_VIF1_BASE + m_VIF1_OFST;
				}
				m_VIF1_STAT ^= STAT1_DBF;

				nImm += CPS2VM::VUMEM1SIZE;
				ExecuteMicro(&CPS2VM::m_VU1, nImm, STAT_VBS1);
				break;
			case 0x4A:
				//MPG
				nAddress += LoadMicro(nAddress, nImm * 8, nNum == 0 ? 256 * 8 : nNum * 8, nUnit == 0 ? CPS2VM::m_pMicroMem0 : CPS2VM::m_pMicroMem1);
				break;
			default:
				assert(0);
				break;
			}
		}
	}
}

uint32 CVIF::UnpackV45(uint32 nSrcAddr, uint32 nDstAddr, uint32 nWriteSize, bool nUsn, uint8 nWL, uint8 nCL, uint8* pVUMem)
{
	unsigned int i, nPackets;
	uint16 nColor;
	uint32 nAddress;

	if(nCL != nWL)
	{
		assert(0);
	}

	nPackets = (nWriteSize / 0x10);
	nAddress = nSrcAddr;

	for(i = 0; i < nPackets; i++)
	{
		nColor = *(uint16*)&CPS2VM::m_pRAM[nAddress];

		*(uint32*)&pVUMem[nDstAddr + 0x0] = ((nColor >>  0) & 0x1F) << 3;
		*(uint32*)&pVUMem[nDstAddr + 0x4] = ((nColor >>  5) & 0x1F) << 3;
		*(uint32*)&pVUMem[nDstAddr + 0x8] = ((nColor >> 10) & 0x1F) << 3;
		*(uint32*)&pVUMem[nDstAddr + 0xC] = ((nColor >> 15) & 0x01) << 7;

		nDstAddr += 0x10;
		nAddress += 0x02;
	}

	//Force word alignment
	if(nAddress & 0x03)
	{
		nAddress += 0x02;
	}

	return nAddress - nSrcAddr;
}

uint32 CVIF::UnpackV432(uint32 nSrcAddr, uint32 nDstAddr, uint32 nWriteSize, bool nUsn, uint8 nWL, uint8 nCL, uint8* pVUMem)
{
	if(nCL != nWL)
	{
		assert(0);
	}

	memcpy(pVUMem + nDstAddr, CPS2VM::m_pRAM + nSrcAddr, nWriteSize);

	return nWriteSize;
}

uint32 CVIF::LoadMicro(uint32 nSrcAddr, uint32 nDstAddr, uint32 nSize, uint8* pMicroMem)
{
	//Check if there's a change
	if(memcmp(pMicroMem + nDstAddr, CPS2VM::m_pRAM + nSrcAddr, nSize) == 0) return nSize;

	CPS2VM::m_VU1.m_pExecMap->InvalidateBlocks();
	memcpy(pMicroMem + nDstAddr, CPS2VM::m_pRAM + nSrcAddr, nSize);
	return nSize;
}

void CVIF::ExecuteMicro(CMIPS* pCtx, uint32 nImm, uint32 nMask)
{
	pCtx->m_State.nPC = nImm;
	m_VPU_STAT |= nMask;
#ifndef VU_DEBUG
	while(m_VPU_STAT & nMask)
	{
		pCtx->Execute(100000);
	}
#endif
}
